/* Definitions of assert-like functions.
   Copyright (C) 2001-2010 Roberto Bagnara <bagnara@cs.unipr.it>
   Copyright (C) 2010-2016 BUGSENG srl (http://bugseng.com)

This file is part of the Parma Polyhedra Library (PPL).

The PPL is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

The PPL is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02111-1307, USA.

For the most up-to-date information see the Parma Polyhedra Library
site: http://bugseng.com/products/ppl/ . */

#include "ppl-config.h"
#include "assertions.hh"
#include <cstdlib>
#include <iostream>

namespace PPL = Parma_Polyhedra_Library;

void
PPL::ppl_unreachable() {
  abort();
}

void
PPL::ppl_unreachable_msg(const char* msg,
                         const char* file, unsigned int line,
                         const char* function) {
  std::cerr << file << ":" << line << ": " << function
            << ": Latent fault detected: " << msg << ".\n";
  abort();
}

void
PPL::ppl_assertion_failed(const char* assertion_text,
                          const char* file, unsigned int line,
                          const char* function) {
  std::cerr << file << ":" << line << ": " << function
            << ": Assertion `" << assertion_text << "' failed.\n";
  abort();
}
